﻿using Phidget22;
using Phidget22.Events;
using Phidget22.ExampleUtils;
using System;
using System.Drawing;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace Gyroscope_Example {
	public partial class Form1 : Form {
		CommandLineOpen openArgs;
		Gyroscope openCh = null;
		formCleaner cleaner;
		Gyroscope gyroscope = null;
		private TemperatureSensor temperatureSensor;
		private ErrorEventBox errorBox;
		private DataGraphBox xGraphBox, yGraphBox, zGraphBox;

		double[] lastMsCount = { 0, 0, 0 };
		bool[] lastMsCountGood = { false, false, false };
		double[] gyroHeading = { 0, 0, 0 }; //degrees

		FastEventHandler<GyroscopeAngularRateUpdateEventArgs> rateEventHandler;

		public Form1() {
			InitializeComponent();
			openArgs = new CommandLineOpen(this);
			cleaner = new formCleaner(this);
		}

		public Form1(Gyroscope ch) {
			InitializeComponent();
			openCh = ch;
			cleaner = new formCleaner(this);
		}

		private void Form1_Load(object sender, EventArgs e) {

			errorBox = new ErrorEventBox(this);
			gyroscope = openCh ?? openArgs.makeChannel<Gyroscope>();

			xGraphBox = new DataGraphBox(this, "°/s");
			yGraphBox = new DataGraphBox(this, "°/s");
			zGraphBox = new DataGraphBox(this, "°/s");

			//register necessary handlers
			gyroscope.Attach += gyroscope_attach;
			gyroscope.Detach += gyroscope_detach;
			gyroscope.Error += gyroscope_error;
			gyroscope.PropertyChange += gyroscope_propertychange;

			// This causes the handler to get called without Invoke
			rateEventHandler = new FastEventHandler<GyroscopeAngularRateUpdateEventArgs>();
			gyroscope.AngularRateUpdate += rateEventHandler.Handler;
			rateEventHandler.Event += angularRate_update;

			// Open the matching TemperatureSensor
			if (openArgs != null) {
				temperatureSensor = openArgs.makeChannel<TemperatureSensor>();
			} else {
				temperatureSensor = new TemperatureSensor();
				temperatureSensor.Channel = gyroscope.Channel;
				temperatureSensor.DeviceSerialNumber = gyroscope.DeviceSerialNumber;
				temperatureSensor.HubPort = gyroscope.HubPort;
				temperatureSensor.IsHubPortDevice = gyroscope.IsHubPortDevice;
				temperatureSensor.IsRemote = gyroscope.IsRemote;
				temperatureSensor.IsLocal = gyroscope.IsLocal;

			}

			temperatureSensor.TemperatureChange += TemperatureSensor_TemperatureChange;
			temperatureSensor.Detach += TemperatureSensor_Detach;

			try {
				gyroscope.Open();
			} catch (PhidgetException ex) { errorBox.addMessage("Error opening device: " + ex.Message); }
		}

		private void Form1_FormClosing(object sender, FormClosingEventArgs e) {

			gyroscope.Attach -= gyroscope_attach;
			gyroscope.Detach -= gyroscope_detach;
			gyroscope.Error -= gyroscope_error;
			gyroscope.PropertyChange -= gyroscope_propertychange;
			gyroscope.AngularRateUpdate -= rateEventHandler.Handler;
			gyroscope.Close();

			temperatureSensor.TemperatureChange -= TemperatureSensor_TemperatureChange;
			temperatureSensor.Detach -= TemperatureSensor_Detach;
			temperatureSensor.Close();
		}

		private void gyroscope_propertychange(object sender, PropertyChangeEventArgs e) {

			switch (e.PropertyName) {
			case "DataInterval":
				if (dataIntervalTrk.TimingMode == PhidgetsControlLibrary.CustomScroll.Modes.INTERVAL) {
					dataIntervalTrk.Value = gyroscope.DataInterval;
					updateGraphDataInterval();
				}
				break;
			case "DataRate":
				if (dataIntervalTrk.TimingMode == PhidgetsControlLibrary.CustomScroll.Modes.RATE) {
					dataIntervalTrk.Value = gyroscope.DataRate;
					updateGraphDataInterval();
				}
				break;
			case "HeatingEnabled":
				heatingEnabledBox.Checked = gyroscope.HeatingEnabled;
				break;
			}
		}
		private void TemperatureSensor_Detach(object sender, DetachEventArgs e) {
			temperatureReading.Text = "Detached";
		}

		private void TemperatureSensor_TemperatureChange(object sender, TemperatureSensorTemperatureChangeEventArgs e) {
			double targetTemp = 50;
			temperatureReading.Text = e.Temperature.ToString() + "°C";
			switch (((TemperatureSensor)sender).DeviceID) {
			case DeviceID.PN_MOT0109:
				targetTemp = 45;
				break;
			}
			if (e.Temperature > (targetTemp - 0.2) && e.Temperature < (targetTemp + 0.2))
				temperatureReading.ForeColor = Color.Green;
			else
				temperatureReading.ForeColor = Color.Red;
		}

		void gyroscope_attach(object sender, Phidget22.Events.AttachEventArgs e) {
			phidgetInfoBox1.FillPhidgetInfo((Phidget)sender);
			Gyroscope attachedDevice = (Gyroscope)sender;

			xGraphBox.SetTitle(DataGraphBox.GenerateTitle("X Axis", (Phidget)attachedDevice));
			yGraphBox.SetTitle(DataGraphBox.GenerateTitle("Y Axis", (Phidget)attachedDevice));
			zGraphBox.SetTitle(DataGraphBox.GenerateTitle("Z Axis", (Phidget)attachedDevice));

			settingsBox.Visible = true;
			outputBox.Visible = true;
			gaugeBox.Visible = true;

			gyroHeading[0] = 0;
			gyroHeading[1] = 0;
			gyroHeading[2] = 0;
			xPosition.Value = 0;
			yPosition.Value = 0;
			zPosition.Value = 0;

			try {
				heatingEnabledBox.Checked = attachedDevice.HeatingEnabled;
				heatingEnabledPanel.Visible = true;
			} catch (PhidgetException ex) {
				if (ex.ErrorCode == ErrorCode.Unsupported)
					heatingEnabledPanel.Visible = false;
			}


			try {//these parameters must be set or the device won't function
				attachedDevice.DataInterval = 32; //setting the interval lower than this is possible but it can cause the GUI to bog down
				dataIntervalTrk.Unit = "ms";
				dataIntervalTrk.initializeRange(attachedDevice.MinDataInterval, attachedDevice.MaxDataInterval);
				dataIntervalTrk.Value = attachedDevice.DataInterval;
			} catch (PhidgetException ex) {
				errorBox.addMessage("Error setting data interval: " + ex.Message);
			}

			guiUpdateTimer.Start();
		}


		void gyroscope_detach(object sender, Phidget22.Events.DetachEventArgs e) {
			guiUpdateTimer.Stop();
			_rate[0] = 0;
			_rate[1] = 0;
			_rate[2] = 0;
			_timestamp = 0;

			phidgetInfoBox1.Clear();
			cleaner.clean();

			settingsBox.Visible = false;
			outputBox.Visible = false;
			gaugeBox.Visible = false;
			dataIntervalTrk.TrueValue = -1;
		}
		void gyroscope_error(object sender, Phidget22.Events.ErrorEventArgs e) {
			errorBox.addMessage(e.Description);

			if (e.Code == ErrorEventCode.Saturation) {
				xGraphBox.addGap();
				yGraphBox.addGap();
				zGraphBox.addGap();
			}
		}

		double[] _rate = { 0, 0, 0 };
		double _timestamp = 0;
		private void guiUpdateTimer_Tick(object sender, EventArgs e) {

			xRatetxt.Text = _rate[0].ToString() + " °/s";
			yRatetxt.Text = _rate[1].ToString() + " °/s";
			zRatetxt.Text = _rate[2].ToString() + " °/s";
			timestamptxt.Text = _timestamp.ToString() + " ms";

			xPosition.Value = (float)gyroHeading[0];
			yPosition.Value = (float)gyroHeading[1];
			zPosition.Value = (float)gyroHeading[2];
		}

		void angularRate_update(object sender, Phidget22.Events.GyroscopeAngularRateUpdateEventArgs e) {
			try {//display the angular rates and calculate the angular positions
				calculateGyroHeading(e, 0);
				calculateGyroHeading(e, 1);
				calculateGyroHeading(e, 2);

				_rate[0] = e.AngularRate[0];
				_rate[1] = e.AngularRate[1];
				_rate[2] = e.AngularRate[2];
				_timestamp = e.Timestamp;

				xGraphBox.addData(e.AngularRate[0], gyroscope.DataInterval, e.Timestamp);
				yGraphBox.addData(e.AngularRate[1], gyroscope.DataInterval, e.Timestamp);
				zGraphBox.addData(e.AngularRate[2], gyroscope.DataInterval, e.Timestamp);

			} catch (PhidgetException ex) {
				errorBox.addMessage("Error reading output: " + ex.Message);
			}
		}

		private void updateGraphDataInterval() {
			try {
				xGraphBox.DataInterval = gyroscope.DataInterval;
				yGraphBox.DataInterval = gyroscope.DataInterval;
				zGraphBox.DataInterval = gyroscope.DataInterval;
			} catch (PhidgetException ex) { errorBox.addMessage("Error updating data interval: " + ex.Message); }
		}

		void calculateGyroHeading(Phidget22.Events.GyroscopeAngularRateUpdateEventArgs data, int index) {
			double gyro = 0;
			gyro = data.AngularRate[index];

			if (Double.IsNaN(gyro))
				gyro = 0;

			if (lastMsCountGood[index]) {//calculate heading
				double timechange = data.Timestamp - lastMsCount[index]; // in ms
				double timeChangeSeconds = (double)timechange / 1000.0;

				gyroHeading[index] += timeChangeSeconds * gyro;

				//ensure proper wrap around
				if (gyroHeading[index] > 180)
					gyroHeading[index] -= 360;
				else if (gyroHeading[index] < -180)
					gyroHeading[index] += 360;
			}

			lastMsCount[index] = data.Timestamp;
			lastMsCountGood[index] = true;
		}

		private void zeroBtn_Click(object sender, EventArgs e) {//zero the gyroscope values
			gyroscope.Zero();
			System.Threading.Thread.Sleep(100);
			gyroHeading[0] = 0;
			gyroHeading[1] = 0;
			gyroHeading[2] = 0;
		}

		private void xGraphBtn_Click(object sender, EventArgs e) {
			if (!xGraphBox.Visible)
				xGraphBox.Show();
		}

		private void yGraphBtn_Click(object sender, EventArgs e) {
			if (!yGraphBox.Visible)
				yGraphBox.Show();
		}

		private void zGraphBtn_Click(object sender, EventArgs e) {
			if (!zGraphBox.Visible)
				zGraphBox.Show();
		}

		private async void heatingEnabledBox_CheckedChanged(object sender, EventArgs e) {
			try {
				heatingEnabledBox.Enabled = false;
				gyroscope.HeatingEnabled = heatingEnabledBox.Checked;
				if (heatingEnabledBox.Checked) {
					try {
						await Task.Run(() => { temperatureSensor.Open(1000); }); //short timeout so that if the temp is already open on another channel we just ignore it
						tempReleaseTimer.Start(); //start checking the temperature.
					} catch (PhidgetException ex) {
						errorBox.addMessage("Error opening temperature channel: " + ex.Message);
					}
				} else if (!heatingEnabledBox.Checked) { //close the channel when heating isn't important to improve speed
					temperatureSensor.Close();
					tempReleaseTimer.Stop();
					temperatureReading.Text = "Unknown";
					temperatureReading.ForeColor = Color.Black;
				}
			} catch (PhidgetException ex) {
				errorBox.addMessage("Error setting heating enabled: " + ex.Message);
			} finally {
				heatingEnabledBox.Enabled = true;
			}
		}
		private void tempReleaseTimer_Tick(object sender, EventArgs e) {
			try {
				if (temperatureSensor.Temperature >= checkSetPoint(gyroscope.DeviceID)) {
					//we have reached the setpoint, we can close the temperature sensor channel now and just display a stable temp message.  this allows for
					//higher data rates to be reached since less data is coming back
					temperatureSensor.Close();
					temperatureReading.ForeColor = Color.Green;
					temperatureReading.Text = "Temp Stablized";
					tempReleaseTimer.Stop();
				}
			} catch (PhidgetException) { }
		}

		private double checkSetPoint(Phidget22.DeviceID id) {
			switch (id) {
			case DeviceID.PN_MOT0109:
				return 45.0;
			case DeviceID.PN_MOT0110:
				return 50.0;
			default:
				return -1;
			}
		}
		private void dataIntervalTrk_KeyOverride(object sender, KeyEventArgs e) {
			e.SuppressKeyPress = true;
			dataIntervalTrk.keyOverride(sender, e);
		}

		private void slider_ValueChangeEvent(object sender, EventArgs e) {
			if (((PhidgetsControlLibrary.CustomScroll)sender).isFixedRange) //datainterval control
			{
				if (dataIntervalTrk.TimingMode == PhidgetsControlLibrary.CustomScroll.Modes.INTERVAL) {
					try {
						double temp = dataIntervalTrk.Value;
						gyroscope.DataInterval = (int)temp;
						if (gyroscope.DataInterval < (int)temp) {
							while (gyroscope.DataInterval != (int)temp) {
								temp++;
								gyroscope.DataInterval = (int)temp;
							}
						} else {
							while (gyroscope.DataInterval != (int)temp) {
								temp--;
								gyroscope.DataInterval = (int)temp;
							}
						}

						dataIntervalTrk.TrueValue = gyroscope.DataInterval;
						dataIntervalTrk.Value = gyroscope.DataInterval;



						updateGraphDataInterval();
					} catch (Exception ex) { errorBox.addMessage("Error setting data interval: " + ex.Message); }
				} else if (dataIntervalTrk.TimingMode == PhidgetsControlLibrary.CustomScroll.Modes.RATE) {
					try {
						gyroscope.DataRate = dataIntervalTrk.Value;
						updateGraphDataInterval();
					} catch (Exception ex) { errorBox.addMessage("Error setting data rate: " + ex.Message); }
				}
			}
		}
		private void dataIntervalTrk_ModeSwitch(object sender, EventArgs e) {
			dataIntervalTrk.SwitchModes();
			if (dataIntervalTrk.TimingMode == PhidgetsControlLibrary.CustomScroll.Modes.RATE) {
				try {
					dataIntervalTrk.isIntValue = false;
					dataIntervalTrk.Unit = "Hz";
					dataIntervalTrk.trkText = gyroscope.DataRate.ToString() + "Hz";
					dataIntervalTrk.Value = gyroscope.DataRate;
				} catch (Exception ex) { errorBox.addMessage("Error setting data rate: " + ex.Message); }

			} else if (dataIntervalTrk.TimingMode == PhidgetsControlLibrary.CustomScroll.Modes.INTERVAL) {
				try {
					dataIntervalTrk.isIntValue = true;
					if (gyroscope.DeviceID != DeviceID.PN_1046)
						dataIntervalTrk.Unit = "ms";
					dataIntervalTrk.trkText = gyroscope.DataInterval.ToString() + "ms";
					dataIntervalTrk.Value = gyroscope.DataInterval;
				} catch (Exception ex) { errorBox.addMessage("Error setting data interval: " + ex.Message); }
			}
			dataIntervalTrk.TrueValue = dataIntervalTrk.Value;
		}
	}
}